import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import soot.Body;
import soot.Local;
import soot.Scene;
import soot.SootClass;
import soot.SootField;
import soot.SootMethod;
import soot.Type;
import soot.RefType;
import soot.Unit;
import soot.Value;
import soot.ValueBox;
import soot.JastAddJ.Stmt;
import soot.jimple.InvokeExpr;
import soot.jimple.ParameterRef;
import soot.jimple.internal.JAssignStmt;
import soot.jimple.internal.JBreakpointStmt;
import soot.jimple.internal.JEnterMonitorStmt;
import soot.jimple.internal.JExitMonitorStmt;
import soot.jimple.internal.JGotoStmt;
import soot.jimple.internal.JIdentityStmt;
import soot.jimple.internal.JIfStmt;
import soot.jimple.internal.JInvokeStmt;
import soot.jimple.internal.JLookupSwitchStmt;
import soot.jimple.internal.JNopStmt;
import soot.jimple.internal.JRetStmt;
import soot.jimple.internal.JReturnStmt;
import soot.jimple.internal.JReturnVoidStmt;
import soot.jimple.internal.JSpecialInvokeExpr;
import soot.jimple.internal.JTableSwitchStmt;
import soot.jimple.internal.JThrowStmt;
import soot.jimple.internal.JimpleLocal;
import soot.toolkits.graph.Block;
import soot.toolkits.graph.BlockGraph;
import soot.toolkits.graph.BriefBlockGraph;
import soot.toolkits.graph.BriefUnitGraph;
import soot.toolkits.graph.DirectedGraph;
import soot.toolkits.graph.DominatorsFinder;
import soot.toolkits.graph.ExceptionalGraph;
import soot.toolkits.graph.MHGDominatorsFinder;
import soot.toolkits.graph.MHGPostDominatorsFinder;
import soot.toolkits.graph.UnitGraph;
import soot.toolkits.scalar.ArraySparseSet;
import soot.toolkits.scalar.FlowSet;
import soot.toolkits.scalar.ForwardFlowAnalysis;
import soot.util.Chain;
import soot.util.cfgcmd.CFGToDotGraph;
import soot.util.dot.DotGraph;
import soot.jimple.toolkits.pointer.util.NativeHelper;

import com.zkh.soot.analysis.*;

/**
 * 
 */

/**
 * @author zkh
 * @param <Directed>
 *
 */
public class IntraCFGFlowAnalysis extends ForwardFlowAnalysis<Unit,FlowSet> {
    FlowSet initialSet;
    FlowSet entryInitialSet;
    boolean return_is_tainted;
    //Map unitToGenerateSet;
    //Map unitToPreserveSet;
    Map<Unit, FlowSet> unitToGenerateSet;
    ArrayList<UserAction> behavior;
    ArrayList<TaintSource> taintSource;

    public IntraCFGFlowAnalysis(DirectedGraph<Unit> graph, FlowSet in) {
    	super(graph);
    	return_is_tainted = false;
    	initialSet = new ArraySparseSet();
    	entryInitialSet = new ArraySparseSet();
    	in.copy(entryInitialSet);
    	doAnalysis();
    }
    
    private Local getLocal(String localName, Chain<Local> lc) {
    	Local l = null;
    	for(Iterator<Local> lci = lc.iterator(); lci.hasNext();) {
    		l = lci.next();
    		if (l.getName().equals(localName))
    			return l;
    	}
    	return l;
    	
    }
    
    private void getUserBehaviorFromFile(String filename) {
		try {
			UserBehaviorInterface userBehavior = new UserBehaviorInterface(filename);			
			this.behavior = userBehavior.getBehavior();
		} catch (Exception e) {
			e.printStackTrace();
		}
    }
    private void getTaintSource(String filename) {
		try {
			TaintSourceInterface _taintSource = new TaintSourceInterface(filename);
			this.taintSource = _taintSource.getTaintSource();
			for(TaintSource action : taintSource) {
				System.out.println("TaintSource " + action);
				this.initialSet.add(action);				
			}
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
    }
    private void buildSequence() {
		SootClass sc;
		SootMethod sm;
		Body sb;
		SootField sf;
		BriefUnitGraph bug;
		Local l;
		Local j;
		Type localType;
		
		if (behavior == null) 
			return;
		for (UserAction action : behavior) {
			sc = Scene.v().getSootClass(action.getClasses());
			sm = sc.getMethodByName(action.getMethod());
			if (action.getWidgetName() == "")
				continue;
			l = getLocal(action.getWidgetName(), sm.getActiveBody().getLocals());
			if (l != null) {
				localType = l.getType();
				if (localType instanceof RefType) {
					//RefType rt = (RefType)
					sc = ((RefType) localType).getSootClass();
					//sm = sc.getMethodByName(action.getAction());
					List<SootMethod> sml = sc.getMethods();
					print(sml.toString());
					sb = sm.getActiveBody();
				}
			}
			sb = sm.getActiveBody();
			bug = new BriefUnitGraph(sb);
			print(sc.toString());
		}
    }
    
	public IntraCFGFlowAnalysis(DirectedGraph<Unit> graph) {
		super(graph);
		//_IntraCFGFlowAnalysis(graph); return;
		a(graph);
	}
	public void a(DirectedGraph<Unit> graph) {
		
		UnitGraph g = (UnitGraph) graph;
		//buildSequence();

		String method_name = g.getBody().getMethod().getName();

		if (//!method_name.equals("main") &&
				!method_name.equals("onModuleLoad") //&&
				//!method_name.equals("sendNameToServer")// &&
				//!method_name.equals("__test__")
				) {
			print("Not a entry (main or OnModuleLoad method), ignore");
			return;
		}
		
		return_is_tainted = false;
		initialSet = new ArraySparseSet();
		String modelPath = Scene.v().getSootClassPath();
		modelPath = modelPath.split(";")[0];
		modelPath = modelPath.replaceAll("war.+", "model");
		getUserBehaviorFromFile(modelPath + "\\userBehavior.xml");
		//"F:\prj\soot\workspace\webtest"
		getTaintSource(modelPath + "\\taintSource.xml");
		entryInitialSet = new ArraySparseSet();
		
		Iterator<UserAction> i = behavior.iterator();
		UserAction ua;
		
		print("@@ Start processing main entry: " + method_name);
		RpcDataFlowAnalysis ana = new RpcDataFlowAnalysis(graph, initialSet, behavior);
		print(ana.getResult().toString());
			
		BriefUnitGraph g2;
		while (i.hasNext()) {
			Unit s = graph.getTails().get(graph.getTails().size()-1);
			initialSet = ana.getResult();
			ua = i.next();
			print("@@ Start processing user actions: " + ua.getWidgetName());
			if ((ua.getCode() == null) || !ua.getCode().hasActiveBody())
				continue;
			
			g2 = new BriefUnitGraph(ua.getCode().getActiveBody());
			ana = new RpcDataFlowAnalysis(g2, initialSet, behavior);
			//print(ana.getResult().toString());
		}
	}
	
	// following is the old version. See previous new version.
	public void _IntraCFGFlowAnalysis(DirectedGraph<Unit> graph) {
		// ___  super(graph);
		
		UnitGraph g = (UnitGraph) graph;
		String method_name = g.getBody().getMethod().getName();
		

		if (//!method_name.equals("main") &&
				!method_name.equals("OnModuleLoad") &&
				!method_name.equals("__test__")) {
			print("Not a entry (main or OnModuleLoad method), ignore");
			return;
		}
		print("@@ Start processing main entry: " + method_name);
		{
			BriefBlockGraph bg = new BriefBlockGraph(g.getBody());
			GetRPCCountInBlock rpc_count = new GetRPCCountInBlock(bg);
			Vector<Vector<Integer>> ret = rpc_count.getRPCCountInBlock(bg.getHeads().get(0));
			
			MHGPostDominatorsFinder pd = new MHGPostDominatorsFinder((DirectedGraph<Block>) bg);
			List<Block> bl = bg.getHeads();
			List<Block> dl = pd.getDominators(bl.get(0));
			Block imd = (Block) pd.getImmediateDominator(bl.get(0));
			CFGToDotGraph ctd = new CFGToDotGraph();
			DotGraph dg = ctd.drawCFG(bg, g.getBody());		
			dg.plot("x:\\test.dot");
		}
		
		return_is_tainted = false;
		initialSet = new ArraySparseSet();
		entryInitialSet = new ArraySparseSet();
		doAnalysis();
	}
	private void dummy_0 (DirectedGraph<Unit> graph) {
        DominatorsFinder df = new MHGDominatorsFinder(graph);
        unitToGenerateSet = new HashMap<Unit, FlowSet>(graph.size() * 2 + 1, 0.7f);

        // pre-compute generate sets
        for(Iterator<Unit> unitIt = graph.iterator(); unitIt.hasNext();){
            Unit s = unitIt.next();
            FlowSet genSet = initialSet.clone();
            
            for(Iterator<Unit> domsIt = df.getDominators(s).iterator(); domsIt.hasNext();){
                Unit dom = domsIt.next();
                for(Iterator<ValueBox> boxIt = dom.getDefBoxes().iterator(); boxIt.hasNext();){
                    ValueBox box = boxIt.next();
                    if(box.getValue() instanceof Local)
                        genSet.add(box.getValue(), genSet);
                }
            }
            
            unitToGenerateSet.put(s, genSet);
        }

        doAnalysis();
	}
	
	protected FlowSet newInitialFlow() {
		return initialSet.clone();
	}
	protected void customizeInitialFlowGraph() {
	
	}
	protected FlowSet entryInitialFlow() {
		return entryInitialSet.clone();
	}
	protected void flowThrough(FlowSet in, Unit unit, FlowSet out) {
		//in.intersection(unitToPreserveSet.get(unit), out);
		//in.intersection(out);
		//out.union(in);		
		print("## "+"new stmt: " + "in= " + in + " ; out = " + out + ", " + unit);
		if (unit instanceof JIdentityStmt) {
			identity_stmt(in, unit, out);
		} else if (unit instanceof JInvokeStmt) {
			invoke_stmt(in, unit, out);
		} else if (unit instanceof JAssignStmt) {
			assign_stmt(in, unit, out);
		} else if (unit instanceof JReturnVoidStmt) {
			return_void_stmt(in, unit, out);
		} else if (unit instanceof JGotoStmt) {
			goto_stmt(in, unit, out);
		} else if (unit instanceof JBreakpointStmt) {
			breakpoint_stmt(in, unit, out);
		} else if (unit instanceof JEnterMonitorStmt) {
			enter_monitor_stmt(in, unit, out);
		} else if (unit instanceof JExitMonitorStmt) {
			exit_monitor_stmt(in, unit, out);
		} else if (unit instanceof JIfStmt) {
			if_stmt(in, unit, out);
		} else if (unit instanceof JLookupSwitchStmt) {
			lookup_switch_stmt(in, unit, out);
		} else if (unit instanceof JNopStmt) {
			nop_stmt(in, unit, out);
		} else if (unit instanceof JRetStmt) {
			ret_stmt(in, unit, out);
		} else if (unit instanceof JReturnStmt) {
			return_stmt(in, unit, out);
		} else if (unit instanceof JTableSwitchStmt) {
			table_switch_stmt(in, unit, out);
		} else if (unit instanceof JThrowStmt) {
			throw_stmt(in, unit, out);
		} else if (unit instanceof JReturnVoidStmt) {
			return_void_stmt(in, unit, out); 
		}

	}
	
	protected void merge(FlowSet in1, FlowSet in2, FlowSet out) {
		in1.union(in2, out);
	}
	
	protected void copy(FlowSet src, FlowSet dest) {
		src.copy(dest);
	}
	private void identity_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("@processing identity_stmt: " + unit);
		in.copy(out);
		//JIdentityStmt s = (JIdentityStmt) unit;
		//s.rightBox. ParameterRef
		JimpleLocal temp = (JimpleLocal) unit.getDefBoxes().iterator().next().getValue();
		/*if (temp.getName().contentEquals("a")) {
			out.add(temp);
		}
		*/
		Value v;
		for(Iterator<ValueBox> i_use = unit.getUseBoxes().iterator(); i_use.hasNext();) {
			v = i_use.next().getValue();			
		    for(Iterator i = out.iterator(); i.hasNext();) {
		    	Object o = i.next();
		    	if (v.equivTo(o))
		    		out.add(temp);
		    }
		}
		/*
		}		*/
	}
	private void invoke_stmt(FlowSet in, Unit unit, FlowSet out) {
		JInvokeStmt stmt = (JInvokeStmt) unit;

		print("@ processing invoke statement: " + unit);
		_InvokeExpr(in, stmt.getInvokeExpr(), out);
		//remove_non_local_feilds(out)
	}
	
	
	//private void _JSpecialInvokeExpr(FlowSet in, JSpecialInvokeExpr expr, FlowSet out) {
	// return true if the invoked method will return tainted data, otherwise return false	
	private boolean _InvokeExpr(FlowSet in, InvokeExpr expr, FlowSet out) {
		Body body;
		SootMethod method;
		BriefUnitGraph graph;
		method =  expr.getMethod();
		FlowSet temp_out = new ArraySparseSet();
		
		
		
		
		if (method.hasActiveBody()) {
			//in.copy(out);
			//////////// Entering a new method //////////////////////
			// processing method call of the same object.
			// should copy tainted class member fields
			
			
			// processing method call of the same class.
			// should copy tainted static class member fields
			//if (this.graph.getBody().get
			// processing the global static variable
			
			
			// processing taint method argument
			Value arg;
			// processing the tainted arguments
			for (int i_arg = 0; i_arg < expr.getArgCount(); i_arg++) {
				arg = expr.getArg(i_arg);
				//if (in.contains(arg)) {					
					ParameterRef pr = new ParameterRef(method.getParameterType(i_arg), i_arg);
					temp_out.add(pr);
				//}
			}

			body = expr.getMethod().getActiveBody();
			graph = new BriefUnitGraph(body);
			IntraCFGFlowAnalysis ana = new IntraCFGFlowAnalysis(graph, temp_out);
			Unit s = graph.getTails().get(graph.getTails().size()-1);
			temp_out = ana.getFlowAfter(s);
			
			//////// post-processing after returned from a method ////////
			// if method belong to the same object, copy tainted object member
			
			// if method belong to the same class, copy tainted static class member
			
			// if method belong to other classes and objects, copy only tainted global variables
			
			ana.getFlowAfter(s).copy(out);
			
			return ana.isReturnTainted();
			//print(out.toString());
			// need to remove non-local variables and non-static class-fields
		} else {
			in.copy(out);
			print("@@ in _InvokeExpr(): no active body for " + expr);
			print(method.getSource().toString());
			return false;
		}
	} // end of _InvokeExpr()
	
	private void assign_stmt(FlowSet in, Unit unit, FlowSet out) {
		JAssignStmt stmt = (JAssignStmt) unit;
		List<ValueBox> def = unit.getDefBoxes();
		Iterator<ValueBox> i_def = def.iterator();
		//FlowSet temp_out = new ArraySparseSet();
		if (stmt.containsInvokeExpr()) { // if invoked another method, follow that method
			print("@ processing assign stmt with InvokeStmt: " + unit);
			if (_InvokeExpr(in, stmt.getInvokeExpr(), out)) {
				for (; i_def.hasNext();) {
					out.add(i_def.next().getValue());
				}
			}
		} else { // normal expr, just check use and add defs to out set 
			List<ValueBox> use = unit.getUseBoxes();
			Value lhs;
			print("@ processing assign stmt: " + unit);
			in.copy(out);
			for (Iterator<ValueBox> i_use = use.iterator(); i_use.hasNext();) {				
				lhs = i_use.next().getValue();
				if (out.contains(lhs)) {  // if any RHS is in in[], the add LHS to out
					for (; i_def.hasNext();) {
						out.add(i_def.next().getValue());
					}
					break;
				}
			}
		}
	}
	
	private void ret_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! JRetStmt is not used and so not supported! " + unit);
		in.copy(out);
	}
	
	// return_stmt is a return statement that returns something.
	private void return_stmt(FlowSet in, Unit unit, FlowSet out) {
		List<ValueBox> use = unit.getUseBoxes();
		print("@ processing return statement: " + unit);
		Value lhs;
		FlowSet temp_out = new ArraySparseSet();
		in.copy(temp_out);
		for (Iterator<ValueBox> i_use = use.iterator(); i_use.hasNext();) {				
			lhs = i_use.next().getValue();
			if (temp_out.contains(lhs)) {  // if any RHS is in in[], the add LHS to out
				return_is_tainted = true;
				break;
			}
		}
		return_void_stmt(temp_out, unit, out);
	}
	
	public boolean isReturnTainted () {
		return return_is_tainted;
	}
	
	// return_void_stmt is a return that returns nothing, just change the execution flow.
	private void return_void_stmt(FlowSet in, Unit unit, FlowSet out) {
		// TODO: should delete local variable from out set
		print("@processing return_void stmt: not fully supported yet! " + unit);
		Value v;
		for(Iterator<Value> i = in.iterator();i.hasNext();) {
			v = i.next();
			if (!(v instanceof ParameterRef) && !(v instanceof Local &&_isLocal(v))) {
				out.add(v);
			}
		}
		//in.copy(out);
	}
	private boolean _isLocal(Value v) {
		UnitGraph g = (UnitGraph) graph;
		Body body = g.getBody();
		Chain<Local> chain = body.getLocals();
		return chain.contains((Local)v);		
	}
	private void breakpoint_stmt(FlowSet in, Unit unit, FlowSet out) {
		in.copy(out);
		print("!! breakpoint_stmt stmt is not supported yet: " + unit);
	}
	private void enter_monitor_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! enter_monitor_stmt stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void exit_monitor_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! exit_monitor_stmt stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void goto_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! goto stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void if_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! if stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void lookup_switch_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! lookup_switch stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void table_switch_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! table switch stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void throw_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! throw stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void nop_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!!@processing nop_stmt: " + unit);
		in.copy(out);
	}
	private void print(String s) {
		System.out.println(s);
	}

}
