/**
 * 
 */
package com.zkh.soot.analysis;

import java.util.Iterator;
import java.util.Vector;

import soot.PatchingChain;
import soot.Unit;
import soot.UnitBox;
import soot.jimple.InvokeExpr;
import soot.toolkits.graph.Block;
import soot.toolkits.graph.BriefBlockGraph;
import soot.toolkits.graph.DirectedGraph;
import soot.toolkits.graph.MHGPostDominatorsFinder;
import soot.toolkits.graph.UnitGraph;
import soot.toolkits.scalar.FlowSet;

/**
 * @author zkh
 * this class will count the number of RPC calls within a basic block.
 * It returns a vector.
 * Each element in the vector has the type of Vector<Integer>,
 * the length of which is the number of different execution paths,
 * and the value of which is the RPC counter on that execution path.
 *
 * This class should be initialized with a BriefBlockGraph.
 * 
 */
public class GetRPCCountInBlock {
	private BriefBlockGraph graph; 
	private Vector<Vector<Integer>> result;
	private Block post_dom = null;
	
	/**
	 * 
	 */
	public GetRPCCountInBlock(BriefBlockGraph graph) {
		this.graph = graph;
		this.result = new Vector<Vector<Integer>>();
	}
	
	public GetRPCCountInBlock(DirectedGraph<Unit> graph, FlowSet in) {
		UnitGraph ug = (UnitGraph) graph;
		this.graph = new BriefBlockGraph(ug.getBody());
		this.result = new Vector<Vector<Integer>>();
	}
	
	/**
	 * This method will call itself recursively to get the PRC call numbers
	 * within the given basic block.
	 * @param Block bb: the basic block to be analyzed.
	 * @return Vector<Vector<Integer>>, the length of the vector equals the number
	 * of different execution path, and each element is the RPC numbers on each path.
	 * Attention that: for each path, the number is a vector too, because that path
	 * can also include multiple execution path itself. 
	 */
	public Vector<Vector<Integer>> getRPCCountInBlock(Block bb) {

		Vector<Vector<Integer>> rpc_return = new Vector<Vector<Integer>>();
		Vector<Vector<Integer>> rpc_succ = new Vector<Vector<Integer>>();		
		MHGPostDominatorsFinder pd_finder = new MHGPostDominatorsFinder((DirectedGraph<Block>) graph);
		
		Block imm_post_dom = (Block) pd_finder.getImmediateDominator(bb);
		if (post_dom == null) {
			post_dom = imm_post_dom;
		}
		Vector<Integer> rpc_local = getCountOfCurrentBlock(bb);
		
		if (bb.equals(post_dom)) {
			Vector<Integer> ret = new Vector<Integer>();
			ret.add(new Integer(0));
			rpc_return.add(ret);
			System.out.println(rpc_return.toString());
			return rpc_return;
		} else	if ( bb.getSuccs().isEmpty()) {//  || bb.getSuccs().contains(imm_post_dom)) { // imm-succ === imm-post-dom) {
			// last block in this path
			//rpc_return.add(new Vector<Integer>(new Integer(0)));
			rpc_return.add(rpc_local );
			System.out.println(rpc_return.toString());
			return rpc_return;
		//} else if ( bb.getSuccs().contains(imm_post_dom)) { // imm-succ === imm-post-dom
		//	rpc_return.add(rpc_local );
		//	return rpc_return;
		} else {
			for(Iterator<Block> i = bb.getSuccs().iterator(); i.hasNext();) {
				rpc_succ=getRPCCountInBlock(i.next());
				// flatten the return vector
				Vector<Integer> rpc_temp = new Vector<Integer>();
				for(Iterator<Vector<Integer>> j=rpc_succ.iterator(); j.hasNext();) {
					rpc_temp.addAll(j.next());				
				}
				for(int k=0; k<rpc_temp.size(); k++) {
					rpc_temp.set(k, new Integer(rpc_temp.get(k)+rpc_local.firstElement().intValue()));
				}
				rpc_return.add(rpc_temp);

			}
			System.out.println(rpc_return.toString());
			return rpc_return;
			
		}
		
	}
	
	/**
	 *  this method will return the RPC count of a current block only, does not
	 * consider the succeeding blocks.
	 * @param bb
	 * @return
	 */
	private Vector<Integer> getCountOfCurrentBlock(Block bb) {
		int count = 0;
		Vector<Integer> ret = new Vector<Integer>();
		PatchingChain<Unit> ul = bb.getBody().getUnits();
		Unit u;

		//for(Iterator<UnitBox> i = bb.getBody().getAllUnitBoxes().iterator(); i.hasNext();) {
		for(Iterator<Unit> i = bb.iterator(); i.hasNext();) {
			u=i.next();//.getUnit();
			System.out.println(u.toString());
			if (isRPCCall(u)) {
				count ++;
			}
		}
		ret.add(new Integer(count));
		System.out.println(ret.toString());		
		return ret;
	}

	/**
	 * this method return true if the given Unit (statement) has a RPC call.
	 * Currently, it is not working properly and will always return true.
	 * @param u
	 * @return
	 */	
	private boolean isRPCCall(Unit u) {
		if (u instanceof InvokeExpr)
			return true;
		else
			return true;
	}
}
