
package com.zkh.soot.analysis;

/**
 * @author zkh
 * this class do the data flow analysis over a given DirectedGraph<Unit>
 * the result can be acquired by call getResult(). That method will return
 * a FlowSet that contains all tainted data.
 * 
 */
	


import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import soot.Body;
import soot.FastHierarchy;
import soot.Local;
import soot.MethodSource;
import soot.RefType;
import soot.Scene;
import soot.SootClass;
import soot.SootMethod;
import soot.Type;
import soot.Unit;
import soot.Value;
import soot.ValueBox;
import soot.JastAddJ.Stmt;
import soot.jimple.Constant;
import soot.jimple.FieldRef;
import soot.jimple.InvokeExpr;
import soot.jimple.ParameterRef;
import soot.jimple.internal.AbstractBinopExpr;
import soot.jimple.internal.AbstractInstanceInvokeExpr;
import soot.jimple.internal.AbstractUnopExpr;
import soot.jimple.internal.JArrayRef;
import soot.jimple.internal.JAssignStmt;
import soot.jimple.internal.JBreakpointStmt;
import soot.jimple.internal.JEnterMonitorStmt;
import soot.jimple.internal.JExitMonitorStmt;
import soot.jimple.internal.JGotoStmt;
import soot.jimple.internal.JIdentityStmt;
import soot.jimple.internal.JIfStmt;
import soot.jimple.internal.JInstanceFieldRef;
import soot.jimple.internal.JInterfaceInvokeExpr;
import soot.jimple.internal.JInvokeStmt;
import soot.jimple.internal.JLookupSwitchStmt;
import soot.jimple.internal.JNopStmt;
import soot.jimple.internal.JRetStmt;
import soot.jimple.internal.JReturnStmt;
import soot.jimple.internal.JReturnVoidStmt;
import soot.jimple.internal.JSpecialInvokeExpr;
import soot.jimple.internal.JTableSwitchStmt;
import soot.jimple.internal.JThrowStmt;
import soot.jimple.internal.JVirtualInvokeExpr;
import soot.jimple.internal.JimpleLocal;
import soot.toolkits.graph.Block;
import soot.toolkits.graph.BlockGraph;
import soot.toolkits.graph.BriefBlockGraph;
import soot.toolkits.graph.BriefUnitGraph;
import soot.toolkits.graph.DirectedGraph;
import soot.toolkits.graph.DominatorsFinder;
import soot.toolkits.graph.ExceptionalGraph;
import soot.toolkits.graph.MHGDominatorsFinder;
import soot.toolkits.graph.MHGPostDominatorsFinder;
import soot.toolkits.graph.UnitGraph;
import soot.toolkits.scalar.ArraySparseSet;
import soot.toolkits.scalar.FlowSet;
import soot.toolkits.scalar.ForwardFlowAnalysis;
import soot.util.Chain;
import soot.util.cfgcmd.CFGToDotGraph;
import soot.util.dot.DotGraph;
import soot.jimple.toolkits.pointer.util.NativeHelper;

import com.zkh.soot.analysis.*;
import com.sidebuster.*;
/**
 * @author zkh
 * @param <Directed>
 *
 */

public class RpcDataFlowAnalysis extends ForwardFlowAnalysis<Unit, FlowSet>{
    FlowSet initialSet;
    FlowSet entryInitialSet;
    FlowSet result;
    boolean return_is_tainted;
    boolean rpc_tainted;
    public static  SidebusterDemo demo;
    //Map unitToGenerateSet;
    //Map unitToPreserveSet;
    Map<Unit, FlowSet> unitToGenerateSet;
    ArrayList<UserAction> behavior;
    Map<Value, Value> pointsTo = new HashMap();

    // this is the main entry that is used normally.
    public RpcDataFlowAnalysis(DirectedGraph<Unit> graph, FlowSet in, ArrayList<UserAction> behavior) {
    	super(graph);
    	return_is_tainted = false;
    	rpc_tainted = false;
    	initialSet = new ArraySparseSet();
    	entryInitialSet = new ArraySparseSet();
    	this.behavior = behavior;
    	in.copy(entryInitialSet);
    	doAnalysis();
    	result = this.getFlowAfter(graph.getTails().get(0));
    }
    
        
    public FlowSet getResult() {
    	return result;
    }
    public boolean metRpcTainted() {
    	return rpc_tainted;
    }
	public RpcDataFlowAnalysis(DirectedGraph<Unit> graph) {
		super(graph);
		
		UnitGraph g = (UnitGraph) graph;
		String method_name = g.getBody().getMethod().getName();
		

		if (//!method_name.equals("main") &&
				!method_name.equals("OnModuleLoad") &&
				!method_name.equals("__test__")) {
			print("Not a entry (main or OnModuleLoad method), ignore");
			return;
		}
		print("            @@ Start processing main entry: " + method_name);
		/*{
			BriefBlockGraph bg = new BriefBlockGraph(g.getBody());
			GetRPCCountInBlock rpc_count = new GetRPCCountInBlock(bg);
			Vector<Vector<Integer>> ret = rpc_count.getRPCCountInBlock(bg.getHeads().get(0));
			
			MHGPostDominatorsFinder pd = new MHGPostDominatorsFinder((DirectedGraph<Block>) bg);
			List<Block> bl = bg.getHeads();
			List<Block> dl = pd.getDominators(bl.get(0));
			Block imd = (Block) pd.getImmediateDominator(bl.get(0));
			CFGToDotGraph ctd = new CFGToDotGraph();
			DotGraph dg = ctd.drawCFG(bg, g.getBody());		
			dg.plot("x:\\test.dot");
		} */
		return_is_tainted = false;
		rpc_tainted = false;
		initialSet = new ArraySparseSet();
		entryInitialSet = new ArraySparseSet();
		doAnalysis();
	}
	
	protected FlowSet newInitialFlow() {
		return initialSet.clone();
	}
	protected void customizeInitialFlowGraph() {
	
	}
	protected FlowSet entryInitialFlow() {
		return entryInitialSet.clone();
	}
	protected void flowThrough(FlowSet in, Unit unit, FlowSet out) {
		if (rpc_tainted) {
			print("## already met a tainted RPC call, return directly");
			return;
		}
		print("## "+"new stmt: " + "in = " + in);
		//print("## "+"new stmt: " + "in = " + in + "; " + unit);
		if (unit instanceof JIdentityStmt) {
			identity_stmt(in, unit, out);
		} else if (unit instanceof JInvokeStmt) {
			invoke_stmt(in, unit, out);
		} else if (unit instanceof JAssignStmt) {
			assign_stmt(in, unit, out);
		} else if (unit instanceof JReturnVoidStmt) {
			return_void_stmt(in, unit, out);
		} else if (unit instanceof JGotoStmt) {
			goto_stmt(in, unit, out);
		} else if (unit instanceof JBreakpointStmt) {
			breakpoint_stmt(in, unit, out);
		} else if (unit instanceof JEnterMonitorStmt) {
			enter_monitor_stmt(in, unit, out);
		} else if (unit instanceof JExitMonitorStmt) {
			exit_monitor_stmt(in, unit, out);
		} else if (unit instanceof JIfStmt) {
			if_stmt(in, unit, out);
		} else if (unit instanceof JLookupSwitchStmt) {
			lookup_switch_stmt(in, unit, out);
		} else if (unit instanceof JNopStmt) {
			nop_stmt(in, unit, out);
		} else if (unit instanceof JRetStmt) {
			ret_stmt(in, unit, out);
		} else if (unit instanceof JReturnStmt) {
			return_stmt(in, unit, out);
		} else if (unit instanceof JTableSwitchStmt) {
			table_switch_stmt(in, unit, out);
		} else if (unit instanceof JThrowStmt) {
			throw_stmt(in, unit, out);
		} else if (unit instanceof JReturnVoidStmt) {
			return_void_stmt(in, unit, out); 
		}
		print("   "+"         " + "out = " + out);

	}
	
	protected void merge(FlowSet in1, FlowSet in2, FlowSet out) {
		in1.union(in2, out);
	}
	
	protected void copy(FlowSet src, FlowSet dest) {
		src.copy(dest);
	}
	private void identity_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("            @processing identity_stmt: " + unit);
		in.copy(out);
		//JIdentityStmt s = (JIdentityStmt) unit;
		//s.rightBox. ParameterRef
		JimpleLocal temp = (JimpleLocal) unit.getDefBoxes().iterator().next().getValue();
		/*if (temp.getName().contentEquals("a")) {
			out.add(temp);
		}
		*/
		Value v;
		Value vLeft;
		Value vRight;
		JIdentityStmt stmt = (JIdentityStmt) unit;
		vLeft = ((ValueBox) stmt.getDefBoxes().get(0)).getValue();
		vRight = ((ValueBox) stmt.getUseBoxes().get(0)).getValue();
		pointsTo.put(vLeft, vRight);
		//print(pointsTo.get(vLeft).toString());
		for(Iterator<ValueBox> i_use = unit.getUseBoxes().iterator(); i_use.hasNext();) {
			v = i_use.next().getValue();			
		    for(Iterator<TaintSource> i = out.iterator(); i.hasNext();) {
		    	Object o = i.next();
		    	if (v.equivTo(o))
		    		out.add(temp);
		    }
		}
		/*
		}		*/
	}
	private void invoke_stmt(FlowSet in, Unit unit, FlowSet out) {
		JInvokeStmt stmt = (JInvokeStmt) unit;

		print("            @ processing invoke statement: " + unit);
		_InvokeExpr(in, stmt.getInvokeExpr(), out);
		//remove_non_local_feilds(out)
	}
	
	private void extractEventHandler(InvokeExpr expr) {
		JimpleLocal jl;
		RefType rt;
		SootClass st;
		SootMethod sm;
		//Body sb;
		
		JVirtualInvokeExpr e = (JVirtualInvokeExpr) expr;
		String sa = e.getMethod().getSignature();
		boolean sab = sa.matches(".*add[a-zA-Z]+Handler.*");
		boolean sac = sa.matches(".*add[a-zA-Z]+Listener.*");
		if (sac) {
			print ("aaa");
		}

		if (sab) {
			sm = null;
			int index = widgetInUserBehaviorModel(e.getBase());
			if ( index >= 0) {
				jl = (JimpleLocal) e.getArg((0));
				rt = (RefType) jl.getType();
				st = rt.getSootClass();
				if (sa.matches(".*ClickHandler.*")) { // button click handler
					sm = st.getMethodByName("onClick");
				} else if (sa.matches(".*KeyUpHandler.*")) { // on key up event
					sm = st.getMethodByName("onKeyUp");
				} else if (sa.matches(".*KeyDownHandler.*")) { // on key down event
					sm = st.getMethodByName("onKeyDown");
				} else if (sa.matches(".*ChangeHandler.*")) {
					sm = st.getMethodByName("onChange");
				} else if (sa.matches(".*SelectionHandler.*")) {
					sm = st.getMethodByName("onSelection");
				}
				updateUserActionCode(index, sm);
				if (sm != null && sm.hasActiveBody()) {
					//sb = sm.getActiveBody();
					print("		get active body for " + sm.toString());
					print("######## " + e.toString() + " #######");
				} else {
					print("		method body not found or no active body for method " + sa);
				}
				
			}
		} else if(sac) { // old version, use add*Listner interface
			sm = null;
			int index = widgetInUserBehaviorModel(e.getBase().toString());
			if ( index >= 0) {
				jl = (JimpleLocal) e.getArg((0));
				rt = (RefType) jl.getType();
				st = rt.getSootClass();
				if (sa.matches(".*ClickListener.*")) { // button click handler
					sm = st.getMethodByName("onClick");
				} else if (sa.matches(".*KeyUpListener.*")) { // on key up event
					sm = st.getMethodByName("onKeyUp");
				} else if (sa.matches(".*KeyDownListener.*")) { // on key down event
					sm = st.getMethodByName("onKeyDown");
				} else if (sa.matches(".*ChangeListener.*")) {
					sm = st.getMethodByName("onChange");
				}
				updateUserActionCode(index, sm);
				if (sm != null && sm.hasActiveBody()) {
					//sb = sm.getActiveBody();
					print("		get active body for " + sm.toString());
					print("######## " + e.toString() + " #######");
				} else {
					print("		method body not found or no active body for method " + sa);
				}
				
			}		
		}
	}
	
	private int widgetInUserBehaviorModel(String widget_name) {
		int index;
		UserAction u;
		for (index=0; index < behavior.size(); index++) {
			u = behavior.get(index);
			if (u.getWidgetName().equals(widget_name)) {
				return index;
			}
		}
		return -1;
	}
	
	private int widgetInUserBehaviorModel(Value v) {
		int index;
		String widget_name;
		UserAction u;
		if ((v instanceof JimpleLocal) && ((JimpleLocal) v).getName().equals("this")) {
			widget_name = ((RefType) ((JimpleLocal) v).getType()).getClassName();
			for (index=0; index < behavior.size(); index++) {
				u = behavior.get(index);
				if (u.getClasses().equals(widget_name)) {
					return index;
				}
			}
		} else {
			widget_name = v.toString();
			for (index=0; index < behavior.size(); index++) {
				u = behavior.get(index);
				if (u.getWidgetName().equals(widget_name)) {
					return index;
				}
			}
		}
		return -1;		
	}
	
	//private void _JSpecialInvokeExpr(FlowSet in, JSpecialInvokeExpr expr, FlowSet out) {
	// return true if the invoked method will return tainted data, otherwise return false	
	
	private boolean isTainted(FlowSet in, InvokeExpr expr) {
		boolean ret = false;
		TaintSource taint;
		Object temp;
		if (!(expr instanceof JVirtualInvokeExpr))
			return false;
		for (Iterator<Object> it = in.iterator(); it.hasNext();) {
			temp = it.next();
			if (!(temp instanceof TaintSource))
				continue;
			taint = (TaintSource) temp;
			if (((JVirtualInvokeExpr) expr).getMethod().getName().equals(taint.getAction()))
				// TODO following if statement should  not be commented out!!!
			//if (((JVirtualInvokeExpr) expr).getBase().toString().equals(taint.getWidgetName()))
					{
				ret = true;
				print("### in _InvokeExpr, get taint-source API " + expr.toString());
				break;
					}			
		}		
		return ret;
	}
	
	private boolean isTainted(FlowSet in, Value v) {
		boolean ret = false;
		Object o;
		for (Iterator<Object> it = in.iterator(); it.hasNext();) {
			o = it.next();
			if (o instanceof Value) {
				if (o.equals(v)) {
					ret = true;
					break;
				} else {
					if (o instanceof JArrayRef) {
						JArrayRef ar = (JArrayRef) o;
						Value arv = ar.getBase();
						Boolean arve = arv.equals(v);
						if (arve) {
							ret = true;
							break;
						}
					}						
				}
			}
		}
		//if (in.contains(v))
		//	ret = true;
		// array, object
		
		return ret;
	}
	
	private boolean isTainted(FlowSet in, String v) {
		boolean ret = false;
		Object o;
		for (Iterator<Object> it = in.iterator(); it.hasNext();) {
			o = it.next();
			if (o instanceof Value) {
				Value t = (Value) o;
				if (t.toString().contains(v)) {
					ret = true;
					break;					
				}
			} else if (o instanceof TaintSource) {
				TaintSource t = (TaintSource) o;
				if (t.getWidgetName().equals(v)) {
					ret = true;
					break;
				}					
			}
		}
		return ret;
	}
	
	private void removeTainted(FlowSet in, String v) {
		Object o;
		for (Iterator<Object> it = in.iterator(); it.hasNext();) {
			o = it.next();
			if (o instanceof Value) {
				Value t = (Value) o;
				if (t.toString().contains(v)) {
					in.remove(o);
					break;					
				}
			} 
		}
	}
	
	private ArrayList<Integer> isTainted(FlowSet in, List<Value> args, boolean rpc_call) {
		ArrayList<Integer> ret = new ArrayList<Integer>();
		Value arg;
		int last = 0;
		if (rpc_call)
			last = 1;
		for (int i_arg = 0; i_arg < args.size()-last; i_arg++) {
			arg = args.get(i_arg);
			if (arg instanceof Constant)
				continue;
			else if (isTainted(in,arg))
				ret.add(i_arg);
		}	
		return ret;
	}
	
	private boolean isRpcCall(InvokeExpr expr) {
		boolean ret = false;		
		if (expr instanceof JInterfaceInvokeExpr && expr.getMethod().getDeclaringClass().isApplicationClass()) {				
			JInterfaceInvokeExpr iie = (JInterfaceInvokeExpr) expr;
			Type iit = iie.getMethodRef().parameterType(iie.getArgCount()-1);
			String iis = iit.toString();
			if (iis.matches("com.google.gwt.user.client.rpc.AsyncCallback")) {
				ret = true;
				print("!! find a RPC call: " + expr.toString());
			}
		}
		return ret;
	}
	
	private SootMethod getRpcCallImpl (InvokeExpr expr) {
		SootMethod ret = null;
		SootMethod method = expr.getMethod();
		String scName = method.getDeclaringClass().toString();
		scName = scName.replace("Async", "");



		SootClass sc = Scene.v().getSootClass(scName);			
		Set s  = Scene.v().getFastHierarchy().getAllImplementersOfInterface(sc);

		if (s.size() == 1) {
			Object [] sa = s.toArray();
			sc = (SootClass) sa[0];
			scName = method.getName();
			ret = sc.getMethodByName(scName);				
		} else {
			ret = null;
			print("Error in getRpcCallImpl");
		}
		return ret;
	}

	private boolean _InvokeExpr(FlowSet in, InvokeExpr expr, FlowSet out) {		
		boolean ret = false;
		Body body;
		SootMethod method;
		BriefUnitGraph graph;
		method =  expr.getMethod();
		FlowSet temp_out = new ArraySparseSet();
		boolean rpc_call;
		boolean selfClear;
		
		if (expr instanceof JVirtualInvokeExpr) {
			extractEventHandler(expr);
			if (isTainted(in, expr)) {
				in.copy(out);
				return true;
			}
		}

		rpc_call = isRpcCall(expr);
		ArrayList<Integer> taintedArgument;
		taintedArgument = isTainted(in,expr.getArgs(), rpc_call);
		//if (rpc_call and taintedArgument.contains(o))
		//	taintedArgument.remove(expr.getArgCount()-1);
		if (rpc_call) {
			if (taintedArgument.size()!=0 && rpc_call) {
				print("%%%% Found tainted RPC call here");
				print("%%%% RPC signature: " + expr.getMethodRef().getSubSignature());
				print("%%%%    tainted arguments: ");
				for (int i = 0; i<taintedArgument.size(); i++)
					print("%%%%    argument " + i + " : " + expr.getArg(i).toString());
 			
				demo.log2("Find a tainted RPC call: " + expr.getMethodRef().getSubSignature());
				String s="";
				for (int i = 0; i<taintedArgument.size(); i++)
					s = s + " argument " + i + " : " + expr.getArg(i).toString();
				demo.log2("      with tainted arguments: " + s +"\n");
			}
			method = getRpcCallImpl(expr);
		}

		//body = method.getActiveBody();
		
		if (method.hasActiveBody()) {
			//in.copy(out);
			//////////// Entering a new method //////////////////////
			// processing method call of the same object.
			// should copy tainted class member fields
			
			
			// processing method call of the same class.
			// should copy tainted static class member fields
			//if (this.graph.getBody().get
			// processing the global static variable
			
			
			// processing taint method argument
			
			// copy tainted data from in to temp_out, except the parameters
			Object iin_temp;
			for (Iterator<Object> iin = in.iterator(); iin.hasNext();) {
				iin_temp = iin.next();
				if (iin_temp instanceof ParameterRef)
					continue;
				else
					temp_out.add(iin_temp);
			}

			//temp_out.union(in);
			Value arg;
			int i_arg;
			// processing the tainted arguments
			//boolean taintedParameter = false;
			for (Iterator<Integer> i = taintedArgument.iterator(); i.hasNext(); ) {
				i_arg = i.next();
			//for (i_arg = 0; i_arg < expr.getArgCount(); i_arg++) {
				//arg = expr.getArg(i_arg);
				//if (isTainted(in, arg)) {					
					ParameterRef pr = new ParameterRef(method.getParameterType(i_arg), i_arg);
					temp_out.add(pr);
					
					//taintedParameter = true;
				//}
			}
						

			//if (taintedParameter) {
				body = method.getActiveBody();
				graph = new BriefUnitGraph(body);
				print("            @@@ recursively processing an invoke" + method.getSubSignature());
				print("%%% Enter " + method.getSubSignature());
				RpcDataFlowAnalysis ana = new RpcDataFlowAnalysis(graph, temp_out, behavior);
				print("%%% Leaving " + method.getSubSignature());
				Unit s = graph.getTails().get(graph.getTails().size()-1);
				temp_out = ana.getFlowAfter(s);
			
				//////// post-processing after returned from a method ////////
				// 	if method belong to the same object, copy tainted object member

				// if method belong to the same class, copy tainted static class member

				// if method belong to other classes and objects, copy only tainted global variables
			
				
				ana.getFlowAfter(s).copy(out);
				ret = ana.isReturnTainted();
				
				if (expr instanceof JSpecialInvokeExpr) { // for contructor, taint this object
					if (isTainted(out, "this")) {
						JSpecialInvokeExpr e = (JSpecialInvokeExpr) expr;
						out.add(e.getBase());
						removeTainted(out, "this");
					}
				}
				
			//} else {
			//	print("            @@ _invoke statement does not contain tainted parameter, ignore!");
			//	return false;
			//}
			//print(out.toString());
			// need to remove non-local variables and non-static class-fields
				
				if (rpc_call) {// process the call-back function after get a RPC call,
					JimpleLocal v0 = (JimpleLocal) expr.getArg(expr.getArgCount()-1);
					RefType p0 = (RefType) v0.getType();
					print(p0.getSootClass().getMethods().toString());
					SootClass sc = p0.getSootClass();
					SootMethod smTemp;
					List<SootMethod> smList = sc.getMethods();
					for(int i = 0; i< sc.getMethodCount(); i++) {
						smTemp = smList.get(i);
						if (smTemp.getName().equals("onFailure")) {
							method = smTemp;
							body = method.getActiveBody();
							graph = new BriefUnitGraph(body);
							print("            @@@ processing RCP call back function" + method.getSubSignature());
							print("%%% Enter " + method.getSubSignature());
							ana = new RpcDataFlowAnalysis(graph, temp_out, behavior);
							print("%%% Leaving " + method.getSubSignature());
							s = graph.getTails().get(graph.getTails().size()-1);
							temp_out = ana.getFlowAfter(s);
						}

						else if (smTemp.getName().equals("onSuccess")) { // &&
								//smTemp.getParameterType(0)) {
							method = smTemp;
							body = method.getActiveBody();
							graph = new BriefUnitGraph(body);
							print("            @@@ processing RCP call back function" + method.getSubSignature());
							print("%%% Enter " + method.getSubSignature());
							if (ret) { // RPC return a tainted data
								ParameterRef pr = new ParameterRef(method.getParameterType(0), 0);
								temp_out.add(pr);
							}
							ana = new RpcDataFlowAnalysis(graph, temp_out, behavior);
							print("%%% Leaving " + method.getSubSignature());
							s = graph.getTails().get(graph.getTails().size()-1);
							temp_out = ana.getFlowAfter(s);
							break;
						}

					}

				
				}

		} else {
			in.copy(out);
			print("            @@ in _InvokeExpr(): no active body for " + expr);
			print("            using pesmistic infering, assuming result is tainted if any parameter is tainted");
			//MethodSource ms = method.getSource();
			//ms.

			//print(method.getSource().toString());
			if (taintedArgument.size() != 0)
				ret = true;
			else if (expr instanceof AbstractInstanceInvokeExpr) {
				AbstractInstanceInvokeExpr tt = (AbstractInstanceInvokeExpr) expr;
				if (isTainted(in, tt.getBase()))
					ret = true;
			}
		}
		return ret;
	} // end of _InvokeExpr()
	
	private void assign_stmt(FlowSet in, Unit unit, FlowSet out) {
		JAssignStmt stmt = (JAssignStmt) unit;
		List<ValueBox> def = unit.getDefBoxes();
		Iterator<ValueBox> i_def = def.iterator();
		
		if (unit.getDefBoxes().get(0).getValue() instanceof JInstanceFieldRef) {
			JInstanceFieldRef r = (JInstanceFieldRef) unit.getDefBoxes().get(0).getValue();
			int index = widgetInUserBehaviorModel(r.getField().getName());
			if ( index >= 0) {
				if (!r.getField().getName().contains("navbar")) {
					return;
				}
				SootMethod sm;
				sm = ((RefType) r.getField().getType()).getSootClass().getMethodByName("onClick");
				updateUserActionCode(index, sm);
			}
		}
		// Testing code
		
		
		
		//FlowSet temp_out = new ArraySparseSet();
		if (stmt.containsInvokeExpr()) { // if invoked another method, follow that method
			print("            @ processing assign stmt with InvokeStmt: " + unit);
			if (_InvokeExpr(in, stmt.getInvokeExpr(), out)) {
				for (; i_def.hasNext();) {
					Value v = i_def.next().getValue();
					if (v instanceof JArrayRef)
						v = ((JArrayRef) v).getBase();
					out.add(v);
				}
			}
		} else { // normal expr, just check use and add defs to out set 
			List<ValueBox> use = unit.getUseBoxes();
			Value rop = stmt.getRightOp();
			boolean fr = stmt.containsFieldRef();
			// TODO: points-to analysis
			/*if (use.size() == 1 && use.get(0).getValue() instanceof JimpleLocal) {
				Value vLeft;
				Value vRight;
				vLeft = ((ValueBox) stmt.getDefBoxes().get(0)).getValue();
				vRight = ((ValueBox) stmt.getUseBoxes().get(0)).getValue();
				pointsTo.put(vLeft, vRight);
			}*/
				
			Value lhs;
			print("            @ processing assign stmt: " + unit);
			in.copy(out);
			for (Iterator<ValueBox> i_use = use.iterator(); i_use.hasNext();) {				
				lhs = i_use.next().getValue();
				if (out.contains(lhs)) {  // if any RHS is in in[], the add LHS to out
					for (; i_def.hasNext();) {
						Value v = i_def.next().getValue();
						if (v instanceof JArrayRef)
							v = ((JArrayRef) v).getBase();
						out.add(v);
					}
					if (lhs instanceof ParameterRef)
						out.remove(lhs);
					break;
				}
			}
		}
	}
	
	private void ret_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! JRetStmt is not used and so not supported! " + unit);
		in.copy(out);
	}
	
	// return_stmt is a return statement that returns something.
	private void return_stmt(FlowSet in, Unit unit, FlowSet out) {
		List<ValueBox> use = unit.getUseBoxes();
		print("            @ processing return statement: " + unit);
		Value lhs;
		FlowSet temp_out = new ArraySparseSet();
		in.copy(temp_out);
		for (Iterator<ValueBox> i_use = use.iterator(); i_use.hasNext();) {				
			lhs = i_use.next().getValue();
			if (temp_out.contains(lhs)) {  // if any RHS is in in[], the add LHS to out
				return_is_tainted = true;
				break;
			}
		}
		return_void_stmt(temp_out, unit, out);
	}
	
	public boolean isReturnTainted () {
		return return_is_tainted;
	}
	
	// return_void_stmt is a return that returns nothing, just change the execution flow.
	private void return_void_stmt(FlowSet in, Unit unit, FlowSet out) {
		// TODO: should delete local variable from out set
		print("            @processing return_void stmt: not fully supported yet! " + unit);
		Value v;
		// TODO: following for-loop is commnet out temperally and should be rewritten later
		// in.copy(out);
		
		Object temp;
		for (Iterator<Object> it = in.iterator(); it.hasNext();) {
			temp = it.next();
			if (temp instanceof TaintSource)
				out.add(temp);
			else {
				v = (Value) temp;
				if (v instanceof ParameterRef || (v instanceof Local &&_isLocal(v)))
					continue;
				else
					out.add(v);
			}
			
		}
		
		//in.copy(out);
	}
	private boolean _isLocal(Value v) {
		UnitGraph g = (UnitGraph) graph;
		Body body = g.getBody();
		Chain<Local> chain = body.getLocals();
		return chain.contains((Local)v);		
	}
	private void breakpoint_stmt(FlowSet in, Unit unit, FlowSet out) {
		in.copy(out);
		print("!! breakpoint_stmt stmt is not supported yet: " + unit);
	}
	private void enter_monitor_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! enter_monitor_stmt stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void exit_monitor_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! exit_monitor_stmt stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void goto_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! goto stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void if_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! if stmt is not supported yet: " + unit);
		JIfStmt stmt = (JIfStmt) unit;
		Value v = stmt.getCondition(); 
		if (v instanceof AbstractBinopExpr) {
			AbstractBinopExpr v2 = (AbstractBinopExpr) v;
			boolean op1Tainted = isTainted(in, v2.getOp1().toString());
			boolean op2Tainted = isTainted(in, v2.getOp2().toString());
			
			if ( op1Tainted || op2Tainted ) {
				print("Find a tainted branch condition!!!");
				demo.log2("Find a tainted branch condition!!!");
				demo.log2("    condition expr is: " + v.toString());
				if (op1Tainted && op2Tainted) {
					demo.log2("     Both operands are tainted: op1 = " + v2.getOp1().toString() + ", op2 = " + v2.getOp2().toString() );
				} else if (op1Tainted) {
					demo.log2("     Only 1st operand is tainted: op1 = " + v2.getOp1().toString());
				} else {
					demo.log2("     Only 2nd operand is tainted: op2 = " + v2.getOp2().toString());
				}
				demo.log2("");
			}
		} else if (v instanceof AbstractUnopExpr) {
				AbstractUnopExpr v2 = (AbstractUnopExpr) v;
			if (isTainted(in, v2.getOp().toString()))
				print("Find a tainted branch condition!!!");
		}
		in.copy(out);
	}
	private void lookup_switch_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! lookup_switch stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void table_switch_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! table switch stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void throw_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!! throw stmt is not supported yet: " + unit);
		in.copy(out);
	}
	private void nop_stmt(FlowSet in, Unit unit, FlowSet out) {
		print("!!            @processing nop_stmt: " + unit);
		in.copy(out);
	}
	private void print(String s) {
		System.out.println(s);
	}

	public ArrayList<UserAction> getBehavior() {
		return behavior;
	}
	private void updateUserActionCode(int index, SootMethod code) {
		UserAction ua;
		String widgetName = behavior.get(index).getWidgetName();
		for(int i=0; i< behavior.size(); i++) {
			ua = behavior.get(i);
			String codeName = code.getName();
			if (ua.getWidgetName().equals(widgetName) && ua.getAction().matches(code.getName())) {
				ua.setCode(code);
				return;
			}
		}
		ua = behavior.get(index);		
		ua.setCode(code);		
	}
}
